import warnings
from reg_tables.utils import compare, align_latex_table, PanelOLSNew
from statsmodels.tools.tools import add_constant
import copy
import io
import pandas as pd
import numpy as np
import re
from varname import argname
from varname.utils import ImproperUseError, UsingExecWarning
from ast import Subscript
from tqdm import tqdm as tq

class Spec():
    """
    Contains specification of regression

    Parameters
    ----------
    data : {np.ndarray, pd.DataFrame}
        Dataset from which 'x' and 'y' variables are going to be sourced from
    y : str
        Name of the column with 'y' variable
    x_vars : {str,list, dict, set, tuple, np.ndarray, pd.core.series.Series}
        Name of the columns with 'x' variables
    entity_effects : bool
        Peform regression with entity effects
    time_effects : bool
        Peform regression with time effects
    other_effects : {str,list, dict, set, tuple, np.ndarray, pd.core.series.Series}
        Category codes to use for any effects that are not entity or time effects. Each variable is treated as an effect.
    cluster_entity : bool
        Cluster standard errors by entity
    cluster_time : bool
        Cluster standard errors by time
    double_cluster : bool
        Cluster standard errors bith by entity and time
    intercept : bool
        Include intercept in the regression
    check_rank : bool
        Check rank during regression
    """

    def __init__(self, 
            data, y, x_vars, 
            entity_effects=False, time_effects=False,
            cluster_entity=False, other_effects=None, cluster_time=False, 
            double_cluster=False, intercept=True,check_rank=True, data_name=None
        ):
        if ((
            (time_effects==True) and
            (entity_effects==True)
            )and\
          (other_effects!=None)):
            raise ValueError('At most two fixed effects are supported.')
        self.data = data
        if data_name != None:
            self.data_name = data_name
        else:
            try:
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=UsingExecWarning)
                    self.data_name = argname('data')
            except ImproperUseError:
                print("Can't retrieve name of dataset. Using default value. Rename the dataset or provide a 'data_name' argument.")
                self.data_name = 'data'
        self.y = y
        if isinstance(x_vars, (list, dict, set, tuple,
                               np.ndarray, pd.core.series.Series)) != True: x_vars = [x_vars]
        self.x_vars = x_vars
        self.entity_effects = entity_effects
        self.time_effects = time_effects
        self.other_effects = other_effects
        self.cluster_entity = cluster_entity
        self.cluster_time = cluster_time
        self.double_cluster = double_cluster
        self.intercept=intercept
        self.check_rank=check_rank
        if (time_effects or entity_effects): intercept=False
        

    def __repr__(self):
        return (f'dataset : {self.data_name}\
                x-vars: {self.x_vars}, y: {self.y},\
                Entity Effects: {self.entity_effects},\
                Time Effects: {self.time_effects},\
                Other Effects: {self.other_effects},\
                Cluster Entity: {self.cluster_entity},\
                Cluster Time: {self.cluster_time},\
                Double Cluster: {self.double_cluster},\
                Intercept: {self.intercept},\
                Check rank: {self.check_rank}')  

    def to_model(self, model):
        """
        Adds the current spec to specified model
        """
        if isinstance(model, Model):
            model.add_spec(self)
        else:
            raise TypeError('Please provide a Model object as argument')

    def run(self):
        """
        Run this regression
        
        Returns
        -------
        PanelEffectsResults
            The panel effects results object.
        
        """
        if self.other_effects != None:             
            reg = PanelOLSNew(
                    self.data[[self.y]],
                    add_constant(self.data[self.x_vars])if self.intercept == True else self.data[self.x_vars], 
                    entity_effects = self.entity_effects, 
                    time_effects = self.time_effects,
                    other_effects = self.data[self.other_effects]
                ).fit(
                cov_type='clustered', 
                cluster_entity=(self.cluster_entity | self.double_cluster),
                cluster_time=(self.cluster_time   | self.double_cluster)
            )
        else:
            reg = PanelOLSNew(
                    self.data[[self.y]],
                    add_constant(self.data[self.x_vars])if self.intercept == True else self.data[self.x_vars], 
                    entity_effects = self.entity_effects, 
                    time_effects = self.time_effects
                ).fit(
                cov_type='clustered', 
                cluster_entity=(self.cluster_entity | self.double_cluster),
                cluster_time=(self.cluster_time   | self.double_cluster)
            )
        return reg
    
                
class Model():
    """
    Contains multiple Spec objects

    Parameters
    ----------
    baseline : Spec
        First regression of the model
    rename_dict : dict
        Rename columns with the variables
    time_entity_effects : bool
        Peform all regressions both with entity and time effects
    """
    def __init__(self, baseline, rename_dict={}, time_entity_effects=False):
        self._rename_dict = rename_dict#add check
        baseline.intercept = True
        self.baseline = baseline
        self.specs = []
        if time_entity_effects:
            for comb in [(False, False), (True, False), (False, True), (True, True)]:
                new_spec = copy.deepcopy(self.baseline)
                new_spec.entity_effects = comb[0]
                new_spec.time_effects = comb[1]
                self.specs.append(new_spec)
        else:
            new_spec = copy.deepcopy(self.baseline)
            self.specs.append(new_spec)
    
    def __repr__(self):
        strr = ''
        for idx, spec in enumerate(self.specs):
            strr = strr+(f'Spec {idx+1}: '+ spec.__repr__()+'\n')
        return strr
    
    def remove_spec (self, idx1, idx2=None):
        """
        Remove regression from the model

        Parameters
        ----------
        idx1 : {float, int}
            Index of the model that needs to be removed 
            (numeration starts from 1)
        idx2 : {float, int}
            If passed a slice of [idx1:idx2] will be removed 
            (numeration starts from 1)
        """
        if idx2 != None:del self.specs[idx1-1:idx2-1] 
        else:
            del self.specs[idx1-1] 
    
    def add_spec(self, **kwargs):
        """
        Add specs to the model
        Parameters
        ----------
        **kwargs:
            kwargs describing the models. Possible arguments :
                data : {np.ndarray, pd.DataFrame}
                    Dataset from which 'x' and 'y' variables are going to be sourced from
                y : str
                    Name of the column with 'y' variable
                x_vars : {str,list, dict, set, tuple, np.ndarray, pd.core.series.Series}
                    Name of the columns with 'x' variables
                entity_effects : bool
                    Peform regression with entity effects
                time_effects : bool
                    Peform regression with time effects
                time_entity_effects : bool
                    Peform regression both with entity and time effects
                other_effects : {str,list, dict, set, tuple, np.ndarray, pd.core.series.Series}
                    Category codes to use for any effects that are not entity or time effects. Each variable is treated as an effect.
                cluster_entity : bool
                    Cluster standard errors by entity
                cluster_time : bool
                    Cluster standard errors by time
                double_cluster : bool
                    Cluster standard errors bith by entity and time
                intercept : bool
                    Include intercept in the regression
                check_rank : bool
                    Check rank during regression
        
        Examples
        --------
        >>> model.add_spec(y='y2', entity_effects=True)
        >>> model.add_spec(y='y2', time_effects=True) 
        """
        new_spec = copy.deepcopy(self.baseline)
        try:
            if isinstance(kwargs[x_vars],(list,dict,set,tuple,np.ndarray,pd.core.series.Series)) != True:
                kwargs[x_vars] = [x_vars] 
        except:pass

        if 'data' in kwargs.keys():
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=UsingExecWarning)
                new_spec.data_name = argname('kwargs[data]')
            if isinstance(new_spec.data_name, Subscript):
                if 'data_name' not in kwargs.keys():
                    new_spec.data_name = 'data'
                    print("Can't retrieve name of dataset. Using default value. Rename the dataset or provide a 'data_name' argument.")

        if (('time_entity_effects' in kwargs) and ('other_effects' in kwargs)) or\
        ((
            ('time_effects' in kwargs) and
            ('entity_effects' in kwargs)
            )\
         and ('other_effects' in kwargs)):
            raise ValueError('At most two fixed effects are supported.')
            

        for key in kwargs: setattr(new_spec, key, kwargs[key])

        if (new_spec.time_effects or new_spec.entity_effects): new_spec.intercept = False

        
        if 'time_entity_effects' in kwargs:
            for comb in [(False, False), (True, False), (False, True), (True, True)]:
                variation = copy.deepcopy(new_spec)
                variation.entity_effects = comb[0]
                variation.time_effects = comb[1]
                variation.intercept = False
                self.specs.append(variation)


        else:self.specs.append(new_spec)


        
    def rename(self, rename_dict):
        """
        Rename the variables in the output table

        Parameters
        ----------
        rename_dict : dict
            Rename columns with the variables
        """
        for key in rename_dict.keys(): self._rename_dict[key] = rename_dict[key]
        
    def run(self,coeff_decimals=2,latex_path=None,
            time_fe_name='Time FEs', entity_fe_name='Entity FEs',
            other_fe_name = 'Other FEs',
            custom_row=None, display_datasets=False,
            rsquared='Inclusive'):
        """
        Run all regressions in the models

        Parameters
        ----------
        coeff_decimals : int
            Display the numbers in the results table 
            with certain number of fraction digits
        latex_path : str
            Write the table in LaTeX format to specified path
        time_fe_name : str
            Name for time fixed effects column
        entity_fe_name : str
            Name for entity fixed effects column
        custom_row : str
            Adds a custom row to the end of the table
        display_datasets : {bool,list}
            Display the names of databases in the results table. 
            If the value is 'True' then use original names of variables.
            Pass a list of strings to define custom names for databases.
        rsquared : str
            Type of R-squared. Default is 'Inclusive'. Other supported types - 'Standard', 'Between',
            'Overall', 'Within'.


        Returns
        -------
        pd.DataFrame
            Table with the results of the regressions
        """
        with warnings.catch_warnings():
            warnings.filterwarnings(
            "ignore",
            message="In a future version of pandas all arguments of concat\
            except for the argument 'objs' will be keyword-only.")
            if custom_row !=None:
                if isinstance (custom_row,list)!=True:
                    print('Custom row is not a list')
            regs = []
            for spec in tq(self.specs, 'Running regressions'):
                regs.append(spec.run())
            

            if rsquared == "Inclusive":
                rsquared_func = lambda reg: reg.rsquared_inclusive
            elif rsquared == "Standard":
                rsquared_func = lambda reg: reg.rsquared
            elif rsquared == "Between":
                rsquared_func = lambda reg: reg.rsquared_between
            elif rsquared == "Overall":
                rsquared_func = lambda reg: reg.rsquared_overall
            elif rsquared == "Within":
                rsquared_func = lambda reg: reg.rsquared_within    

            R2s  = list(map(rsquared_func, regs))
            regs = compare(regs, stars=True, precision='tstats')
            csv = regs.summary.as_csv()
            tab = pd.read_csv(io.StringIO(csv), skiprows=1)
            

            other_eff_dict = {}
            for idx, spec in enumerate(self.specs):
                if spec.other_effects != None:
                    other_eff_dict[idx] = spec.other_effects
            if other_eff_dict != {}:
                rows = tab.shape[0]
                cols = tab.shape[1]
                last_row = tab[rows-1:rows].copy()
                tab = tab[:rows-1]
                tab.loc[tab.shape[0]] = ['Other Effects'] + ['']*(cols-1)
                tab = pd.concat([tab,last_row])
                for key, value in other_eff_dict.items():
                    if isinstance (value, list):
                        renamed_items = []
                        for item in value:
                            if item in self._rename_dict.keys():
                                item = self._rename_dict[item]
                            renamed_items.append(item)
                        tab.iat[tab.shape[0]-2,int(key)+1] = ', '.join(renamed_items)
                    else:  
                        if value in self._rename_dict.keys():
                                    value = self._rename_dict[value]
                        tab.iat[tab.shape[0]-2,int(key)+1] = value
            tab = tab.set_index([tab.columns[0]])
            col_dict = dict(zip(tab.columns.to_list(), 
                              list(map(lambda x:'('+str(int(x.replace(' ','').replace('Model',''))+1)+')',
                                   tab.columns.to_list())))
            )
            coeff_borders = []
            observ = int()
            r2 = int()
            const = int()
            for idx, x in enumerate(tab.index):
                if 'No. Observations' in x:observ = idx
                if 'const' in x:const = idx
                if re.match('R-squared    ',x) != None:
                    r2 = idx
                if '===' in x:coeff_borders.append(idx)
            
            tab.rename(index={tab.index[observ]:'Observations', tab.index[const]:'Intercept'}, columns=col_dict, inplace=True)
            tab.loc['Observations'] = ["{0:0,.0f}".format(float(x)) for x in tab.loc['Observations']]
            try:coeffs = tab[coeff_borders[0]+1:coeff_borders[1]].copy()
            except:coeffs = tab[coeff_borders[0]+1:-1].copy()
            if coeff_decimals != None:
                def change_decimals(cell):
                    try:
                        if '(' in cell:
                            sub_str = float(re.search('\(-?[0-9]*\.[0-9]*' ,cell)[0][1:])
                            return '('+re.sub('\(-?[0-9]*\.[0-9]*', f'{sub_str:.{coeff_decimals}f}', cell)
                        else:
                            sub_str = float(re.search('^-?[0-9]*\.[0-9]*', cell)[0])
                            return re.sub('^-?[0-9]*\.[0-9]*', f'{sub_str:.{coeff_decimals}f}', cell)
                    except TypeError:
                        return cell
                coeffs = coeffs.applymap(change_decimals)
                s = "{0:0."+str(coeff_decimals)+"f}"
                R2s  = [ s.format(x) for x in R2s ]
            else:
                R2s  = [ "{0:0.4f}".format(x) for x in R2s ]

            if const != 0:
                coeffs = pd.concat([coeffs[2:], coeffs[0:2]])
            coeffs_dict = {}
            
            for idx,name in enumerate(coeffs.index):
                if re.sub('[ \t]+$','',name) in self._rename_dict.keys():
                    coeffs_dict[name] = self._rename_dict[re.sub('[ \t]+$', '', name)]
            coeffs.rename(index=coeffs_dict, inplace=True)
            final = pd.concat([tab.head(1), coeffs])
            for idx,name in enumerate(final.iloc[0]):
                if re.sub('[ \t]+$','', name) in self._rename_dict.keys():
                   final.iat[0,idx] = self._rename_dict[re.sub('[ \t]+$','', name)]

            # Add spacing
            final = pd.concat([final.iloc[:1], pd.DataFrame(index=[' ']), final.iloc[1:]])
            final = pd.concat([final, pd.DataFrame(index=[' '])])

            
            for line in [observ,r2]:
                final = pd.concat([final, tab[line:].head(1)])

            # Inclusive R2s (including fixed effects)
            final.iloc[-1] = R2s
            
            effects = pd.DataFrame(index=[time_fe_name, entity_fe_name, other_fe_name])
            some_effects = False
            for column in tab.columns:
                for x in tab[column]:
                    if re.search('Time', str(x))!=None: effects.loc[time_fe_name,column]='Yes'; some_effects = True
                    if re.search('Entity', str(x))!=None: effects.loc[entity_fe_name,column]='Yes'; some_effects = True
            if other_eff_dict != {}:
                for x in tab.iloc[tab.shape[0]-2,:].index:
                    effects.at[other_fe_name, x] = tab.iloc[tab.shape[0]-2,:][x]
                some_effects = True
            else:
                effects = effects[:2]
            if some_effects: final=pd.concat([final,effects])

            if display_datasets != False:
                if display_datasets == True:
                    data_info = pd.DataFrame(data=[[spec.data_name for spec in self.specs]]
                                            , columns=final.columns, index=['Dataset'])
                else:
                    data_info = pd.DataFrame(data=[display_datasets]
                        , columns=final.columns, index=['Dataset'])
                final = pd.concat([final, data_info])
            if custom_row != None:
                custom = pd.DataFrame(index=[custom_row[0]])
                for idx,item in enumerate(custom_row[1:]):
                    custom.at[custom_row[0],final.columns[idx]]=item
                final = pd.concat([final,custom])
            final.fillna('', inplace=True)
            final.drop(index=' ', inplace=True)
            if latex_path != None:
                latex_string = final.style.to_latex(column_format = 'l'+ ((final.shape[1])*'c'), hrules=True)
                # latex_string = re.sub('(?<=\{tabular\}\{l)(.*?)(?=\})',
                #                     'c'*len(re.search('(?<=\{tabular\}\{l)(.*?)(?=\})',
                #                     final.style.to_latex())[0]),final.style.to_latex())
                # latex_string = re.sub('{lcccc}\n','{lcccc}\n\\\\toprule\n{}', latex_string)
                # latex_string = re.sub('\nD','\n\\\midrule\nD', latex_string)
                # latex_string = re.sub('\n\\\end{tabular}\n','\n\\\\bottomrule\n\\\end{tabular}\n', latex_string)
                latex_string = align_latex_table(latex_string).replace('\\\\ \n',' \\\\\n')
                with open(latex_path, 'w') as f:
                    f.write(latex_string)   
            return final


from __future__ import annotations
from linearmodels.panel.results import PanelModelComparison,PanelModelResults
from linearmodels.panel.model import PanelOLS
from typing import Dict, List, Union, cast
from linearmodels.compat.statsmodels import Summary
from pandas import  Series, concat
from linearmodels.shared.io import add_star, pval_format
from linearmodels.iv.results import default_txt_fmt, stub_concat, table_concat
from linearmodels.panel.covariance import (
    setup_covariance_estimator,
)
from linearmodels.panel.data import PanelData
from linearmodels.panel.results import (
    PanelEffectsResults,
)
from linearmodels.panel.utility import (
    AbsorbingEffectWarning,
    absorbing_warn_msg,
    check_absorbed,
    not_absorbed,
)

from linearmodels.shared.hypotheses import (
    WaldTestStatistic,
)

from linearmodels.typing import (
    Float64Array,
    IntArray,
)
from scipy.linalg import lstsq as sp_lstsq
from statsmodels.iolib.summary import SimpleTable
import numpy as np
from pandas import  DataFrame, Series
import numpy as np
import re

def _lstsq(
    x: Float64Array, y: Float64Array, rcond: float | None = None
) -> tuple[Float64Array, Float64Array, int, Float64Array]:
    if rcond is None:
        eps = np.finfo(np.float64).eps
        cond = float(max(x.shape) * eps)
    else:
        cond = rcond
    return sp_lstsq(x, y, cond=cond, lapack_driver="gelsy")

def align_latex_table(table):
    lines = table.split('\n')
    start_idx = lines.index(list(filter(lambda x: '&' in x, lines))[0])
    try:
        end_idx = lines.index(list(filter(lambda x: 'Entity FEs ' in x, lines))[0])+1
    except IndexError:
        end_idx = lines.index(list(filter(lambda x: 'R-squared ' in x, lines))[0])+1
    columns = lines[start_idx].count('&')+1
    max_len = [0] * columns
    f_lines = []
    #changhing lines formatting  
    for idx, line in enumerate(lines):
        if (idx < start_idx) | (idx > end_idx):
            f_lines.append(line)
            continue

        if '&' not in line:
            f_lines.append(line)
            continue
        
        items = line.split('&')
        new_line = []
        for item_idx, item in enumerate(items):
            if '\\\\' in item:
                item = item.replace('\\\\','')
            try:
                item_0 = re.search(r"\S", item).span()[0]
            except Exception:
                new_line.append(' ')
                continue
            item_1 = re.search(r"\S\s*$", item).span()[0]+1
            new_item = item[item_0:item_1]
            new_line.append(' '+new_item+' ')
        f_lines.append('&'.join(new_line))


    for idx, line in enumerate(f_lines):
        if (idx < start_idx) | (idx > end_idx):
            continue

        if '&' not in line:
            continue
        
        items = line.split('&')
        for item_idx,item in enumerate(items):
            length = len(item)
            if item_idx == len(max_len)-1:
                length += 4
            if length> max_len[item_idx]:
                max_len[item_idx] = length

    new_lines = []    
    for idx, line in enumerate(f_lines):
        if (idx < start_idx) | (idx > end_idx):
            new_lines.append(line)
            continue

        if '&' not in line:
            new_lines.append(line)
            continue
        
        items = line.split('&')
        new_line = []
        for item_idx, item in enumerate(items):


            if len(item) == max_len[item_idx]:
                if item_idx == len(max_len)-1:
                    item = item[:-4] + '\\\\'
                new_line.append(item)
            else:
                if item_idx == len(max_len)-1:
                    item = item + ' '*(max_len[item_idx] - len(item) -4) + '\\\\'
                    new_line.append(item)
                else:
                    item = item + ' '*(max_len[item_idx] - len(item))
                    
                    new_line.append(item)
        
        new_lines.append('&'.join(new_line))
    new_table = "\n".join(new_lines)
    return new_table



def _str(v: float) -> str:
            if np.isnan(v):
                return " "

            av = abs(v)
            digits = 0

            if av != 0:
                digits = int(np.ceil(np.log10(av)))

            if digits > 4 or digits <= -4:
                return "{0:8.4g}".format(v)
            else:
                return "{:.20f}".format(v)


class InvalidDataset(Exception):
    "Raised when there are issues with the dataset"
    pass

class PanelOLSNew(PanelOLS):
    def fit(
        self,
        *,
        use_lsdv: bool = False,
        use_lsmr: bool = False,
        low_memory: bool | None = None,
        cov_type: str = "unadjusted",
        debiased: bool = True,
        auto_df: bool = True,
        count_effects: bool = True,
        **cov_config: bool | float | str | IntArray | DataFrame | PanelData,
    ) -> PanelEffectsResults:
        """
        Estimate model parameters

        Parameters
        ----------
        use_lsdv : bool
            Flag indicating to use the Least Squares Dummy Variable estimator
            to eliminate effects.  The default value uses only means and does
            note require constructing dummy variables for each effect.
        use_lsmr : bool
            Flag indicating to use LSDV with the Sparse Equations and Least
            Squares estimator to eliminate the fixed effects.
        low_memory : {bool, None}
            Flag indicating whether to use a low-memory algorithm when a model
            contains two-way fixed effects. If `None`, the choice is taken
            automatically, and the low memory algorithm is used if the
            required dummy variable array is both larger than then array of
            regressors in the model and requires more than 1 GiB .
        cov_type : str
            Name of covariance estimator. See Notes.
        debiased : bool
            Flag indicating whether to debiased the covariance estimator using
            a degree of freedom adjustment.
        auto_df : bool
            Flag indicating that the treatment of estimated effects in degree
            of freedom adjustment is automatically handled. This is useful
            since clustered standard errors that are clustered using the same
            variable as an effect do not require degree of freedom correction
            while other estimators such as the unadjusted covariance do.
        count_effects : bool
            Flag indicating that the covariance estimator should be adjusted
            to account for the estimation of effects in the model. Only used
            if ``auto_df=False``.
        **cov_config
            Additional covariance-specific options.  See Notes.

        Returns
        -------
        PanelEffectsResults
            Estimation results

        Examples
        --------
        >>> from linearmodels import PanelOLS
        >>> mod = PanelOLS(y, x, entity_effects=True)
        >>> res = mod.fit(cov_type="clustered", cluster_entity=True)

        Notes
        -----
        Three covariance estimators are supported:

        * "unadjusted", "homoskedastic" - Assume residual are homoskedastic
        * "robust", "heteroskedastic" - Control for heteroskedasticity using
          White's estimator
        * "clustered` - One- or two-way clustering.  Configuration options are:

          * ``clusters`` - Input containing 1 or 2 variables.
            Clusters should be integer valued, although other types will
            be coerced to integer values by treating as categorical variables
          * ``cluster_entity`` - Boolean flag indicating to use entity
            clusters
          * ``cluster_time`` - Boolean indicating to use time clusters

        * "kernel" - Driscoll-Kraay HAC estimator. Configurations options are:

          * ``kernel`` - One of the supported kernels (bartlett, parzen, qs).
            Default is Bartlett's kernel, which is produces a covariance
            estimator similar to the Newey-West covariance estimator.
          * ``bandwidth`` - Bandwidth to use when computing the kernel.  If
            not provided, a naive default is used.
        """

        weighted = np.any(self.weights.values2d != 1.0)

        if use_lsmr:
            y, x, ybar, y_effects, x_effects = self._lsmr_path()
        elif use_lsdv:
            y, x, ybar, y_effects, x_effects = self._slow_path()
        else:
            low_memory = (
                self._choose_twoway_algo() if low_memory is None else low_memory
            )
            if not weighted:
                y, x, ybar = self._fast_path(low_memory=low_memory)
                y_effects = np.array([0.0])
                x_effects = np.zeros(x.shape[1])
            else:
                y, x, ybar, y_effects, x_effects = self._weighted_fast_path(
                    low_memory=low_memory
                )

        neffects = 0
        drop_first = self.has_constant
        if self.entity_effects:
            neffects += self.dependent.nentity - drop_first
            drop_first = True
        if self.time_effects:
            neffects += self.dependent.nobs - drop_first
            drop_first = True
        if self.other_effects:
            assert self._other_effect_cats is not None
            oe = self._other_effect_cats.dataframe
            for c in oe:
                neffects += oe[c].nunique() - drop_first
                drop_first = True

        if self.entity_effects or self.time_effects or self.other_effects:
            if not self._drop_absorbed:
                check_absorbed(x, [str(var) for var in self.exog.vars])
            else:
                # TODO: Need to special case the constant here when determining which
                #  to retain since we always want to retain the constant if present
                retain = not_absorbed(x, self._constant, self._constant_index)
                if not retain:
                    raise ValueError(
                        "All columns in exog have been fully absorbed by the included"
                        " effects. This model cannot be estimated."
                    )
                if len(retain) != x.shape[1]:
                    drop = set(range(x.shape[1])).difference(retain)
                    dropped = ", ".join([str(self.exog.vars[i]) for i in drop])
                    import warnings

                    warnings.warn(
                        absorbing_warn_msg.format(absorbed_variables=dropped),
                        AbsorbingEffectWarning,
                        stacklevel=2,
                    )
                    x = x[:, retain]
                    # Update constant index loc
                    if self._constant:
                        assert isinstance(self._constant_index, int)
                        self._constant_index = int(
                            np.argwhere(np.array(retain) == self._constant_index)
                        )

                    # Adjust exog
                    self.exog = PanelData(self.exog.dataframe.iloc[:, retain])
                    x_effects = x_effects[retain]

        params = _lstsq(x, y, rcond=None)[0]
        nobs = self.dependent.dataframe.shape[0]
        df_model = x.shape[1] + neffects
        df_resid = nobs - df_model
        # Check clusters if singletons were removed
        cov_config = self._setup_clusters(cov_config)
        if auto_df:
            count_effects = self._determine_df_adjustment(cov_type, **cov_config)
        extra_df = neffects if count_effects else 0
        cov = setup_covariance_estimator(
            self._cov_estimators,
            cov_type,
            y,
            x,
            params,
            self.dependent.entity_ids,
            self.dependent.time_ids,
            debiased=debiased,
            extra_df=extra_df,
            **cov_config,
        )

        weps = y - x @ params
        eps = weps
        _y = self.dependent.values2d
        _x = self.exog.values2d
        if weighted:
            eps = (_y - y_effects) - (_x - x_effects) @ params
            if self.has_constant:
                # Correction since y_effects and x_effects @ params add mean
                w = self.weights.values2d
                eps -= (w * eps).sum() / w.sum()
        index = self.dependent.index
        fitted = DataFrame(_x @ params, index, ["fitted_values"])
        idiosyncratic = DataFrame(eps, index, ["idiosyncratic"])
        eps_effects = _y - fitted.values

        sigma2_tot = float(np.squeeze(eps_effects.T @ eps_effects) / nobs)
        sigma2_eps = float(np.squeeze(eps.T @ eps) / nobs)
        sigma2_effects = sigma2_tot - sigma2_eps
        rho = sigma2_effects / sigma2_tot if sigma2_tot > 0.0 else 0.0

        resid_ss = float(np.squeeze(weps.T @ weps))
        if self.has_constant:
            mu = ybar
        else:
            mu = np.array([0.0])
        total_ss = float(np.squeeze((y - mu).T @ (y - mu)))
        r2 = 1 - resid_ss / total_ss if total_ss > 0.0 else 0.0

        root_w = cast(Float64Array, np.sqrt(self.weights.values2d))
        y_ex = root_w * self.dependent.values2d
        mu_ex = 0
        if (
            self.has_constant
            or self.entity_effects
            or self.time_effects
            or self.other_effects
        ):
            mu_ex = root_w * ((root_w.T @ y_ex) / (root_w.T @ root_w))
        total_ss_ex_effect = float(np.squeeze((y_ex - mu_ex).T @ (y_ex - mu_ex)))
        r2_ex_effects = (
            1 - resid_ss / total_ss_ex_effect if total_ss_ex_effect > 0.0 else 0.0
        )

        res = self._postestimation(params, cov, debiased, df_resid, weps, y, x, root_w)
        ######################################
        # Pooled f-stat
        ######################################
        if self.entity_effects or self.time_effects or self.other_effects:
            wy, wx = root_w * self.dependent.values2d, root_w * self.exog.values2d
            df_num, df_denom = (df_model - wx.shape[1]), df_resid
            if not self.has_constant:
                # Correction for when models does not have explicit constant
                wy -= root_w * _lstsq(root_w, wy, rcond=None)[0]
                wx -= root_w * _lstsq(root_w, wx, rcond=None)[0]
                df_num -= 1
            weps_pooled = wy - wx @ _lstsq(wx, wy, rcond=None)[0]
            resid_ss_pooled = float(np.squeeze(weps_pooled.T @ weps_pooled))
            num = (resid_ss_pooled - resid_ss) / df_num

            denom = resid_ss / df_denom
            if denom == 0:
                raise InvalidDataset('There is a problem with your dataset. This error may indicate that the dependent variable in your data is constant.') 
            stat = num / denom
            f_pooled = WaldTestStatistic(
                stat,
                "Effects are zero",
                df_num,
                df_denom=df_denom,
                name="Pooled F-statistic",
            )
            res.update(f_pooled=f_pooled)
            effects = DataFrame(
                eps_effects - eps,
                columns=["estimated_effects"],
                index=self.dependent.index,
            )
        else:
            effects = DataFrame(
                np.zeros_like(eps),
                columns=["estimated_effects"],
                index=self.dependent.index,
            )

        res.update(
            dict(
                df_resid=df_resid,
                df_model=df_model,
                nobs=y.shape[0],
                residual_ss=resid_ss,
                total_ss=total_ss,
                wresids=weps,
                resids=eps,
                r2=r2,
                entity_effects=self.entity_effects,
                time_effects=self.time_effects,
                other_effects=self.other_effects,
                sigma2_eps=sigma2_eps,
                sigma2_effects=sigma2_effects,
                rho=rho,
                r2_ex_effects=r2_ex_effects,
                effects=effects,
                fitted=fitted,
                idiosyncratic=idiosyncratic,
            )
        )

        return PanelEffectsResults(res)


class PanelModelComparisonNew(PanelModelComparison):

    @property
    def summary(self) -> Summary:
        """
        Model estimation summary.

        Returns
        -------
        Summary
            Summary table of model estimation results

        Notes
        -----
        Supports export to csv, html and latex  using the methods ``summary.as_csv()``,
        ``summary.as_html()`` and ``summary.as_latex()``.
        """


        
        smry = Summary()
        models = list(self._results.keys())
        title = "Model Comparison"
        stubs = [
            "Dep. Variable",
            "Estimator",
            "No. Observations",
            "Cov. Est.",
            "R-squared",
            "R-Squared (Within)",
            "R-Squared (Between)",
            "R-Squared (Overall)",
            "F-statistic",
            "P-value (F-stat)",
        ]
        dep_name = {}
        for key in self._results:
            dep_name[key] = self._results[key].model.dependent.vars[0]
        dep_name = Series(dep_name)

        vals = concat(
            [
                dep_name,
                self.estimator_method,
                self.nobs,
                self.cov_estimator,
                self.rsquared,
                self.rsquared_within,
                self.rsquared_between,
                self.rsquared_overall,
                self.f_statistic,
            ],
            axis=1,
        )
        vals = [[i for i in v] for v in vals.T.values]
        vals[2] = [str(v) for v in vals[2]]
        for i in range(4, len(vals)):
            f = _str
            if i == 9:
                f = pval_format
            vals[i] = [f(v) for v in vals[i]]

        params = self.params
        precision = getattr(self, self._precision)
        pvalues = np.asarray(self.pvalues)
        params_fmt = []
        params_stub = []
        for i in range(len(params)):
            formatted_and_starred = []
            for v, pv in zip(params.values[i], pvalues[i]):
                formatted_and_starred.append(add_star(_str(v), pv, self._stars))
            params_fmt.append(formatted_and_starred)

            precision_fmt = []
            for v in precision.values[i]:
                v_str = _str(v)
                v_str = "({0})".format(v_str) if v_str.strip() else v_str
                precision_fmt.append(v_str)
            params_fmt.append(precision_fmt)
            params_stub.append(params.index[i])
            params_stub.append(" ")

        vals = table_concat((vals, params_fmt))
        stubs = stub_concat((stubs, params_stub))

        all_effects = []
        for key in self._results:
            res = self._results[key]
            effects = getattr(res, "included_effects", [])
            all_effects.append(effects)

        neffect = max(map(len, all_effects))
        effects = []
        effects_stub = ["Effects"]
        for i in range(neffect):
            if i > 0:
                effects_stub.append("")
            row = []
            for j in range(len(self._results)):
                effect = all_effects[j]
                if len(effect) > i:
                    row.append(effect[i])
                else:
                    row.append("")
            effects.append(row)
        if effects:
            vals = table_concat((vals, effects))
            stubs = stub_concat((stubs, effects_stub))

        txt_fmt = default_txt_fmt.copy()
        txt_fmt["data_aligns"] = "r"
        txt_fmt["header_align"] = "r"
        table = SimpleTable(
            vals, headers=models, title=title, stubs=stubs, txt_fmt=txt_fmt
        )
        smry.tables.append(table)
        prec_type = self._PRECISION_TYPES[self._precision]
        smry.add_extra_txt(["{0} reported in parentheses".format(prec_type)])
        return smry


def compare(
    results: Union[List[PanelModelResults], Dict[str, PanelModelResults]],
    *,
    precision: str = "tstats",
    stars: bool = False,
) -> PanelModelComparisonNew:
    """
    Compare the results of multiple models

    Parameters
    ----------
    results : {list, dict}
        Set of results to compare.  If a dict, the keys will be used as model
        names.
    precision : {"tstats","std_errors", "std-errors", "pvalues"}
        Estimator precision estimator to include in the comparison output.
        Default is "tstats".
    stars : bool
        Add stars based on the p-value of the coefficient where 1, 2 and
        3-stars correspond to p-values of 10%, 5% and 1%, respectively.

    Returns
    -------
    PanelModelComparison
        The model comparison object.
    """
    return PanelModelComparisonNew(results, precision=precision, stars=stars)
